from typing import Any

from dagster import In, Nothing, Out
from qdk.base import BaseComponent


class BaseLoader(BaseComponent):
    compute_function = "load"
    tags = {
        "kind": "loader",
    }
    input_defs = {"after": In(Nothing)}
    output_defs = {"out": Out(Any)}

    @classmethod
    def load(cls) -> Any:
        """
        Overwrite this function to load the required object.
        """
        raise NotImplementedError(
            'Make sure you added a "load" function to the component'
        )



import os
from dataclasses import dataclass


@dataclass
class S3Connection:
    aws_access_key_id: str = os.environ.get("AWS_ACCESS_KEY_ID")
    aws_access_secret_key: str = os.environ.get("AWS_SECRET_ACCESS_KEY")
    aws_endpoint_url: str = os.environ.get("AWS_ENDPOINT_URL")



from typing import Optional, Tuple

from sqlalchemy import create_engine
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.sql.schema import MetaData


def create_sql_engine(database_url: str) -> Tuple[Engine, MetaData]:
    """
    Create a connection to the database using the supplied database url
    """
    engine = create_engine(database_url)
    meta = MetaData()
    meta.bind = engine
    meta.reflect(views=True)

    return (engine, meta)


def get_primary_key(conn: Connection, table_name: str) -> Optional[str]:
    primary_key = conn.execute(
        f"""
        SELECT DISTINCT tco.constraint_name
        FROM information_schema.table_constraints tco
                 JOIN information_schema.key_column_usage kcu
                      ON kcu.constraint_name = tco.constraint_name
                          AND kcu.constraint_schema = tco.constraint_schema
                          AND kcu.constraint_name = tco.constraint_name
        WHERE kcu.table_name = '{table_name}'
          AND constraint_type = 'PRIMARY KEY';
        """
    ).first()

    if primary_key:
        return primary_key[0]
    else:
        return None



import pickle
from typing import Any

from dagster import Field, Out, StringSource
from dagster_aws.s3 import s3_resource
from qdk.loader.base import BaseLoader


class S3PickleLoader(BaseLoader):
    output_defs = {"result": Out(Any)}
    config_schema = {
        "bucket": Field(
            StringSource,
            is_required=True,
        ),
        "location": Field(
            StringSource,
            is_required=True,
        ),
    }

    required_resource_keys = {
        "s3",
    }

    @classmethod
    def load(cls, s3: s3_resource, bucket: str, location: str) -> Any:
        response = s3.get_object(Bucket=bucket, Key=location)
        body_string = response["Body"].read()
        object = pickle.loads(body_string)

        return object



from typing import Union

import dask.dataframe as dd
import pandas as pd
from dagster import Field, StringSource
from pandas.io.sql import SQLTable
from qdk.exporter.base import BaseExporter
from qdk.utils.sqlalchemy import create_sql_engine, get_primary_key
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.engine import Connection


class DatabaseExporter(BaseExporter):
    config_schema = {
        "connection_uri": Field(
            StringSource,
            is_required=True,
            description="The SqlAlchemy connection uri. (postgresql://user:password@host:port/database)",
        ),
        "table_name": Field(
            str,
            is_required=True,
            description="The name of the table the data will be inserted to.",
        ),
    }

    @classmethod
    def upsert(
        cls,
        table: SQLTable,
        conn: Connection,
        keys: list,
        data_iter: zip,
    ):
        # Get the name of the primary key, used for upserting
        primary_key_name = get_primary_key(conn, table.name)

        if not primary_key_name:
            raise ValueError(
                "The table you are inserting into does not have a primary key."
            )

        upsert_args = {
            "constraint": primary_key_name,
        }

        # Loop through the data and upsert each row
        for data in data_iter:
            data = {k: data[i] for i, k in enumerate(keys)}
            upsert_args["set_"] = data
            insert_stmt = insert(cls.meta.tables[table.name]).values(**data)
            upsert_stmt = insert_stmt.on_conflict_do_update(**upsert_args)
            conn.execute(upsert_stmt)

    @classmethod
    def export(
        cls,
        df: Union[pd.DataFrame, dd.DataFrame],
        connection_uri: str,
        table_name: str,
    ) -> None:
        """This component uploads data to a database. You can upsert the data
        if you are using a Postgres database.

        Args:
            df (Union[pd.DataFrame, dd.DataFrame]): The dataframe you want to upload
            connection_uri (str): The sqlalchemy uri of the database
            table_name (str): The name of the table you want to upload to
        """
        engine, meta = create_sql_engine(connection_uri)

        # Bind meta to the class, because we cannot pass it
        # to the upsert function
        cls.meta = meta

        with engine.connect() as conn:
            df.to_sql(
                table_name,
                con=conn,
                if_exists="append",
                method=cls.upsert,
                index=False,
            )



from typing import List, Union

import dask.dataframe as dd
import nltk
import pandas as pd
import spacy
from dagster import Field
from qdk.transform.base import BaseTransformer
from qdk.utils.text import clean_text


class NounPhraseTransformer(BaseTransformer):
    noun_phrase_regex = "NP: {<NOUN.*|ADJ>*<NOUN.*>}"
    remove_pos = {"PUNCT"}
    config_schema = {
        "text_columns": Field(
            list,
            default_value=["text"],
            description="Columns to tokenize",
        ),
        "spacy_model": Field(
            str,
            default_value="nl_core_news_lg",
            description="The name of the spacy language model (nl_core_news_lg)",
        ),
        "keep_tokens": Field(
            bool,
            default_value=False,
            description="Whether to keep all tokens, or only noun phrases",
        ),
    }

    @classmethod
    def extract_noun_phrases(
        cls,
        text: str,
        nlp: str,
        keep_tokens: bool,
    ) -> List[str]:
        """This function takes a string as input and outputs a list of
        noun phrase candidates:

        "Dit is een zin met tekst" -> ["zin", "tekst"]

        Args:
            text (str): The text you would like to transform.
            nlp (str): A spacy language model. Used for POS tagging.
            keep_tokens (bool): Whether to keep all tokens, or only noun phrases

        Returns:
            List[str]: The noun phrases found in the text.
        """
        # Clean the text before extracting phrases
        cleaned_text = clean_text(text=text)

        # Use spacy to parse and tokenize the text
        doc = nlp(cleaned_text)

        # Only retrieve the text and part of speech tag
        doc = [(token.text.lower(), token.pos_) for token in doc]

        # Use the nltk regex parser to extract different parts of speech
        parser = nltk.RegexpParser(cls.noun_phrase_regex)
        tree = parser.parse(doc)

        # Buffer for all noun phrases and extra tokens
        tokens = []

        # Only select the deepest subtree, this prevents duplicate tokens
        for subtree in tree.subtrees(lambda t: t.height() == tree.height()):
            # The items in the subtree are either a word or a noun phrase subtree
            for phrase in subtree:
                # If it is a noun phrase subtree, join it and add it to the tokens buffer
                if isinstance(phrase, nltk.tree.Tree):
                    noun_phrases = phrase.leaves()
                    tokens.append(" ".join(word for word, _ in noun_phrases))
                # Otherwise keep the word if the user requests it
                elif keep_tokens:
                    word, pos = phrase
                    if pos not in cls.remove_pos:
                        tokens.append(word)

        return tokens

    @classmethod
    def transform(
        cls,
        df: Union[pd.DataFrame, dd.DataFrame],
        text_columns: List[str] = ["text"],
        spacy_model: str = "nl_core_news_lg",
        keep_tokens: bool = False,
    ) -> Union[pd.DataFrame, dd.DataFrame]:
        nlp = spacy.load(
            spacy_model,
            disable=[
                "parser",
                "ner",
                "lemmatizer",
                "textcat",
            ],
        )

        # Use dask to parallelize the keyword extraction over the workers
        if type(df) == dd.DataFrame:
            for text_column in text_columns:
                df[f"_nouns_{text_column}"] = df.map_partitions(
                    lambda _df: _df[text_column].apply(
                        cls.extract_noun_phrases, args=(nlp, keep_tokens)
                    )
                )
        # Pandas apply function
        else:
            for text_column in text_columns:
                df[f"_nouns_{text_column}"] = df[text_column].apply(
                    cls.extract_noun_phrases, args=(nlp, keep_tokens)
                )

        # Return the dataframe with the keywords
        return df



from typing import Dict, List, Tuple, Union

import dask.dataframe as dd
import pandas as pd
from dagster import Field, In, Out
from dask_ml.model_selection import train_test_split
from qdk.dagster_types import DataFrameType, SeriesType
from qdk.transform.base import BaseTransformer

import yake


class TrainTestTransformer(BaseTransformer):
    input_defs = {"df": In(DataFrameType)}
    output_defs = {
        "X_train": Out(DataFrameType),
        "X_test": Out(DataFrameType),
        "y_train": Out(SeriesType),
        "y_test": Out(SeriesType),
    }
    config_schema = {
        "target_column": Field(
            str,
            is_required=True,
            description="The column to use as target",
        ),
        "test_size": Field(
            float,
            default_value=0.2,
            description="The proportion of the dataset to include in the test split",
        ),
        "shuffle": Field(
            bool,
            default_value=True,
            description="Whether or not to shuffle the data before splitting",
        ),
    }

    @staticmethod
    def _extract_keywords(
        text, extractor: yake.KeywordExtractor
    ) -> List[Tuple[str, float]]:
        """Extracts keywords from a string using a yake KeywordExtractor instance.

        Args:
            text (str): The text you want to extract the keywords from.
            extractor (yake.KeywordExtractor): The initialized yake keyword extractor.

        Returns:
            List[Tuple[str, float]]: Returns a list with keyword, score tuples. The scores indicates keyword relevence (the lower the better).
        """
        return extractor.extract_keywords(text)

    @classmethod
    def transform(
        cls,
        df: Union[pd.DataFrame, dd.DataFrame],
        target_column: str,
        test_size: float = 0.2,
        shuffle: bool = False,
    ) -> Tuple[
        Union[pd.DataFrame, dd.DataFrame],
        Union[pd.DataFrame, dd.DataFrame],
        Union[pd.Series, dd.Series],
        Union[pd.Series, dd.Series],
    ]:
        """
        Args:
            df (Union[pd.DataFrame, dd.DataFrame]): The dataframe to split.
            target_column (str): The column to use as target.

        Returns:
            Tuple[
                Union[pd.DataFrame, dd.DataFrame],
                Union[pd.DataFrame, dd.DataFrame],
                Union[pd.Series, dd.Series],
                Union[pd.Series, dd.Series]
            ]: Returns a tuple with the train and test dataframes, the train and test target series.
        """
        X = df.drop(target_column, axis=1)
        y = df[target_column]

        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=test_size, shuffle=shuffle
        )

        return (X_train, X_test, y_train, y_test)



from typing import Optional


def clean_text(text: Optional[str]) -> str:
    """Perform basic text cleaning
    1. Normalize special characters
    2. Remove special quotes

    Args:
        text (Optional[str]): The text you want to clean

    Returns:
        str: The cleaned text
    """
    if not text:
        return ""

    text = text.replace("è", "e")
    text = text.replace("é", "e")
    text = text.replace("ê", "e")
    text = text.replace("ë", "e")
    text = text.replace("ì", "i")
    text = text.replace("í", "i")
    text = text.replace("î", "i")
    text = text.replace("ï", "i")
    text = text.replace("ò", "o")
    text = text.replace("ó", "o")
    text = text.replace("ô", "o")
    text = text.replace("õ", "o")
    text = text.replace("ö", "o")
    text = text.replace("ù", "u")
    text = text.replace("ú", "u")
    text = text.replace("û", "u")
    text = text.replace("ü", "u")
    text = text.replace("ç", "c")
    text = text.replace("‘", "")
    text = text.replace("’", "")
    text = text.replace("©", "")

    return text



import multiprocessing
from typing import Union

import dask.dataframe as dd
import pandas as pd
from dagster import Field, In, Out
from gensim.models.word2vec import Word2Vec
from qdk.dagster_types import DataFrameType, ModelType
from qdk.training.base import BaseTrainer


class Word2VecTrainer(BaseTrainer):
    tags = {
        "kind": "word2vec",
    }
    input_defs = {"X": In(DataFrameType)}
    output_defs = {"model": Out(ModelType)}
    config_schema = {
        "token_column": Field(
            str,
            default_value="_tokens_text",
        ),
        "vector_size": Field(
            int,
            default_value=100,
        ),
        "window": Field(
            int,
            default_value=5,
        ),
        "min_count": Field(
            int,
            default_value=5,
        ),
        "negative": Field(
            int,
            default_value=5,
        ),
    }

    @classmethod
    def train(
        cls,
        X: Union[pd.DataFrame, dd.DataFrame],
        token_column: str = "_tokens_text",
        vector_size: int = 100,
        window: int = 5,
        min_count: int = 5,
        negative: int = 5,
    ) -> Word2Vec:
        # Train the Word2Vec model
        word2vec_model = Word2Vec(
            sentences=X[token_column],
            vector_size=vector_size,
            window=window,
            min_count=min_count,
            negative=negative,
            workers=multiprocessing.cpu_count(),
        )

        return word2vec_model



from typing import List, Tuple, Union

import dask.dataframe as dd
import pandas as pd
from dagster import Enum, EnumValue, Field
from qdk.transform.base import BaseTransformer

import yake


class YakeTransformer(BaseTransformer):
    config_schema = {
        "text_columns": Field(
            list,
            default_value=["text"],
            description="Columns to tokenize",
        ),
        "language": Field(
            Enum("Language", [EnumValue("en"), EnumValue("nl")]),
            default_value="nl",
            description="The language of the input text.",
        ),
        "n_grams": Field(
            int, default_value=1, description="The maximum length of keyword n-grams."
        ),
        "dedup_limit": Field(
            float, default_value=0.9, description="The deduplication limit."
        ),
        "n_keywords": Field(
            int, default_value=10, description="How many keywords to extract."
        ),
    }

    @staticmethod
    def _extract_keywords(
        text, extractor: yake.KeywordExtractor
    ) -> List[Tuple[str, float]]:
        """Extracts keywords from a string using a yake KeywordExtractor instance.

        Args:
            text (str): The text you want to extract the keywords from.
            extractor (yake.KeywordExtractor): The initialized yake keyword extractor.

        Returns:
            List[Tuple[str, float]]: Returns a list with keyword, score tuples. The scores indicates keyword relevence (the lower the better).
        """
        return extractor.extract_keywords(text)

    @classmethod
    def transform(
        cls,
        df: Union[pd.DataFrame, dd.DataFrame],
        text_columns: List[str] = ["text"],
        language: str = "nl",
        n_grams: int = 1,
        dedup_limit: float = 0.9,
        n_keywords: int = 10,
    ) -> Union[pd.DataFrame, dd.DataFrame]:
        """Use the YAKE method to extract keywords from the text column in the dataframe.

        Args:
            df (Union[pd.DataFrame, dd.DataFrame]): The dataframe to perform the keyword extraction on.
            language (str, optional): Which base language to use for the keyword extraction. Defaults to "nl".
            n_grams (int, optional): The max length of the keywords. Defaults to 1.
            dedup_limit (float, optional): The deduplication limit. Defaults to 0.9.

        Returns:
            Union[pd.DataFrame, dd.DataFrame]: Returns a dataframe with a "_keywords" column.
        """
        # Create the yake keywork extractor
        extractor = yake.KeywordExtractor(
            lan=language,
            n=n_grams,
            dedupLim=dedup_limit,
            top=n_keywords,
        )

        # Use dask to parallelize the keyword extraction over the workers
        if type(df) == dd.DataFrame:
            for text_column in text_columns:
                df[f"_keywords_{text_column}"] = df.map_partitions(
                    lambda _df: _df[text_column].apply(
                        cls._extract_keywords, args=(extractor,)
                    )
                )
        # Pandas apply function
        else:
            for text_column in text_columns:
                df[f"_keywords_{text_column}"] = df[text_column].apply(
                    cls._extract_keywords, args=(extractor,)
                )

        # Return the dataframe with the keywords
        return df



from typing import Any, Dict, List, Optional, Sequence, Set

from dagster import Field, In, OpDefinition, OpExecutionContext, Out, Output


class BaseComponent:
    input_defs: Dict[str, In] = {}
    output_defs: Dict[str, Out] = {}
    config_schema: Dict[str, Field] = {}
    required_resource_keys: Set[str] = set()
    tags: Dict[str, str] = {}

    @classmethod
    def _compute_hook(cls, step_context: OpExecutionContext, inputs: Dict[str, Any]):
        """The compute hook that runs the actual compute function.
            Needs to overridden by the descendant classes.

        Args:
            step_context (OpExecutionContext): The Dagster execution context.
            inputs (Dict[str, Any]): The inputs that are supplied to the op as a dictionary.

        Raises:
            NotImplementedError: The compute hook is not implemented by default.
        """
        # Load the required resources defined in the required_resource_keys attribute
        resources = {
            resource_key: resource
            for resource_key, resource in step_context.resources._asdict().items()
            if resource_key in cls.required_resource_keys
        }

        # Run the compute function ("cls.compute_function" needs to be defined by the children)
        # The inputs of the previous op, resources and op config are injected into the compute function
        results = getattr(cls, cls.compute_function)(
            **inputs,
            **resources,
            **step_context.op_config,
        )

        # If the results are not a sequence, wrap it in a list
        if not isinstance(results, Sequence):
            results = [results]

        # Zip the results with the output definitions and yield the results
        for result, output_def in zip(results, cls.output_defs):
            yield Output(result, output_def)

    @classmethod
    def to_op(cls, name: str, description: Optional[str] = None) -> OpDefinition:
        """Generates a op definition for this component.

        Args:
            name (str): The name of the op.
            input_defs (List[In], optional): The list of Dagster input values. Defaults to [].
            output_defs (List[Out], optional): The list of Dagster output values. Defaults to [].

        Returns:
            OpDefinition: The operator that is used by Dagster to execute this component.
        """
        return OpDefinition(
            name=name,
            description=description,
            ins=cls.input_defs,
            outs=cls.output_defs,
            required_resource_keys=cls.required_resource_keys,
            config_schema=cls.config_schema,
            compute_fn=cls._compute_hook,
            tags=cls.tags,
        )



from mlflow.pyfunc import PyFuncModel, load_model
from mlflow import get_tracking_uri
from mlflow.tracking.fluent import Run
from pandas import DataFrame


class MLFlowRun:
    def __init__(self, run: Run, model=None):
        self.run = run
        self._model = model

    @property
    def run_id(self):
        return self.run.info.run_id

    @property
    def run_url(self):
        return f"{get_tracking_uri()}/#/experiments/{self.run.info.experiment_id}/runs/{self.run_id}"

    @property
    def run_url_local(self):
        return f"http://localhost:5000/#/experiments/{self.run.info.experiment_id}/runs/{self.run_id}"

    @property
    def model_flavors(self):
        return self.model.metadata.flavors.keys()

    @property
    def model_uri(self) -> str:
        return f"runs:/{self.run_id}/model"

    @property
    def input_schema(self) -> DataFrame:
        return DataFrame(self.model.metadata.get_input_schema().to_dict())

    @property
    def output_schema(self) -> DataFrame:
        return DataFrame(self.model.metadata.get_output_schema().to_dict())

    @property
    def model(self) -> PyFuncModel:
        # If a model was supplied to the constructor, use that.
        if self._model:
            return self._model
        # Otherwise, load the model from mlflow using its run id.
        else:
            return load_model(self.model_uri)



from typing import Any, Dict, List, Optional, Union

import dask.dataframe as dd
import pandas as pd
from dagster import Field, Noneable, Permissive, StringSource
from qdk.loader.dataframe import DataFrameLoader


class SqlTableLoader(DataFrameLoader):
    config_schema = {
        "connection_uri": Field(
            StringSource,
            description="The SqlAlchemy connection uri. (postgresql://user:password@host:port/database)",
        ),
        "table": Field(
            str,
            description="The database table to load the data from.",
        ),
        "use_dask": Field(
            bool,
            default_value=False,
            description="Whether to load the dataframe using Dask.",
        ),
        "index_column": Field(
            Noneable(str),
            default_value=None,
            description="Which database column should be used as dataframe index.",
        ),
        "drop_na": Field(
            Noneable([str]),
            default_value=None,
            description="Whether to drop rows with missing values. ('column', or ['col_a', 'col_b'])",
        ),
        "drop_duplicates": Field(
            Noneable([str]),
            default_value=None,
            description="Whether to drop duplicates in certain columns. ('column', or ['col_a', 'col_b'])",
        ),
        "load_params": Field(
            Permissive({}),
            description="Extra parameters that get passed to the loading function.",
        ),
    }

    @classmethod
    def load(
        cls,
        connection_uri: str,
        table: str,
        use_dask: bool = False,
        index_column: Optional[str] = None,
        drop_na: List[str] = None,
        drop_duplicates: List[str] = None,
        load_params: Dict[str, Any] = {},
    ) -> Union[pd.DataFrame, dd.DataFrame]:
        # Choose which framework to use for data loading
        framework = dd if use_dask else pd

        # Read the sql table using the sqlalchemy uri
        df = framework.read_sql_table(
            table,
            connection_uri,
            index_col=index_column,
            **load_params,
        )

        # Post process the data that is coming in
        df = cls.post_process(
            df,
            drop_na=drop_na,
            drop_duplicates=drop_duplicates,
            compute_ddf=False,
        )

        return df



# from dagster import (
#     ModeDefinition,
#     OpExecutionContext,
#     PipelineExecutionResult,
#     execute_pipeline,
#     pipeline,
# )
# from dask.dataframe import DataFrame as DaskDataFrame
# from pandas import DataFrame
# from qdk import DataFrameLoader, YakeTransformer, qdk_io_manager

# from ..utils.dataframe import CSVStoreDataFrame

# loader_op = DataFrameLoader.to_op(
#     name="loader",
# )

# keyword_op = YakeTransformer.to_op(
#     name="keyword",
# )


# @pipeline(mode_defs=[ModeDefinition(resource_defs={"io_manager": qdk_io_manager})])
# def nlp_pipeline():
#     df = loader_op()
#     keyword = keyword_op(df)


# def run_pipeline(tmpdir: str, use_dask: bool) -> OpExecutionContext:
#     # Create a temporary dataframe and store it in the temp directory
#     df_storer = CSVStoreDataFrame()
#     df_path = df_storer.store_dataframe(tmpdir)

#     # Execute the pipeline
#     result = execute_pipeline(
#         nlp_pipeline,
#         run_config={
#             "ops": {
#                 "loader": {
#                     "config": {
#                         "uri": df_path,
#                         "use_dask": use_dask,
#                     }
#                 },
#                 "keyword": {
#                     "config": {
#                         "language": "en",
#                         "text_columns": ["description"],
#                     },
#                 },
#             }
#         },
#     )

#     return result


# def test_nlp_pipeline(tmpdir):
#     # Execute the pipeline
#     result = run_pipeline(
#         tmpdir,
#         use_dask=False,
#     )

#     result_df = result.output_for_solid("keyword", "df")

#     assert result.success
#     assert isinstance(result, PipelineExecutionResult)
#     assert isinstance(result_df, DataFrame)


# def test_nlp_dask_pipeline(tmpdir):
#     # Execute the pipeline
#     result = run_pipeline(
#         tmpdir,
#         use_dask=True,
#     )

#     result_df = result.output_for_solid("keyword", "df")

#     assert result.success
#     assert isinstance(result, PipelineExecutionResult)
#     assert isinstance(result_df, DaskDataFrame)



from typing import Optional, Tuple, Union

import dask.dataframe as dd
import pandas as pd
from dagster import In, Nothing
from dagster import Optional as DagsterOptional
from dagster import Out
from qdk.base import BaseComponent
from qdk.dagster_types import DataFrameType, ModelType, SeriesType

from sklearn.base import BaseEstimator


class BaseTrainer(BaseComponent):
    compute_function = "train"
    tags = {
        "kind": "training",
    }
    input_defs = {
        "X": In(DataFrameType),
        "y": In(DataFrameType),
        "model": In(DagsterOptional[SeriesType]),
    }
    output_defs = {
        "model": Out(ModelType),
        "data": Out(DataFrameType, is_required=False),
    }

    @classmethod
    def train(
        cls,
        X: Union[pd.DataFrame, dd.DataFrame],
        model: Optional[BaseEstimator] = None,
        y: Optional[Union[pd.Series, dd.Series]] = None,
    ) -> Tuple[BaseEstimator, Optional[Union[pd.DataFrame, dd.DataFrame]]]:
        raise NotImplementedError(
            'Make sure you added a "train" function to the component'
        )



from gensim.models import Word2Vec
from qdk.models import Word2VecModel


def test_word2vec_model():
    X = [["a", "b"]]

    # Instanciate the model
    w2v_model = Word2VecModel(
        vector_size=5,
        min_count=1,
    )

    # Make sure there is no fitted model yet
    assert w2v_model.model == None

    # Fit the model
    w2v_model.fit(X)

    # Make sure the model is fitted now
    assert isinstance(w2v_model.model, Word2Vec)

    # Test the fit_transform functionality
    word_vectors = w2v_model.fit_transform(X)

    assert len(word_vectors) == 1
    assert len(word_vectors[0]) == 5


def test_word2vec_model_mean_vector():
    X = [["a", "b"]]

    # Instanciate the model
    w2v_model = Word2VecModel(
        vector_size=5,
        min_count=1,
        mean_vector=False,
    )

    word_vectors = w2v_model.fit_transform(X)

    assert len(word_vectors[0]) == 2
    assert len(word_vectors[0][0]) == 5



from typing import Union

import dask.dataframe as dd
import pandas as pd
from dagster import In
from qdk.dagster_types import DataFrameType, SeriesType
from qdk.training.base import BaseTrainer

from sklearn.base import BaseEstimator


class SklearnTrainer(BaseTrainer):
    input_defs = {
        "X": In(DataFrameType),
        "y": In(SeriesType),
        "model": In(BaseEstimator),
    }

    required_resource_keys = {"mlflow"}

    @classmethod
    def train(
        cls,
        X: Union[pd.DataFrame, dd.DataFrame],
        y: Union[pd.Series, dd.Series],
        model: BaseEstimator,
    ):
        return model.fit(X, y)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Author: Chinmaya Pancholi <chinmayapancholi13@gmail.com>
# Copyright (C) 2017 Radim Rehurek <radimrehurek@seznam.cz>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html

"""Scikit learn interface for :class:`~gensim.models.word2vec.Word2Vec`.
Follows scikit-learn API conventions to facilitate using gensim along with scikit-learn.
Examples
--------
.. sourcecode:: pycon
    >>> from gensim.test.utils import common_texts
    >>> from gensim.sklearn_api import W2VTransformer
    >>>
    >>> # Create a model to represent each word by a 10 dimensional vector.
    >>> model = W2VTransformer(size=10, min_count=1, seed=1)
    >>>
    >>> # What is the vector representation of the word 'graph'?
    >>> wordvecs = model.fit(common_texts).transform(['graph', 'system'])
    >>> assert wordvecs.shape ==2, 10)
"""
from typing import List, Optional

import numpy as np
from gensim.models import Word2Vec
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError


class Word2VecModel(TransformerMixin, BaseEstimator):
    """Base Word2Vec module, wraps :class:`~gensim.models.word2vec.Word2Vec`.
    For more information please have a look to `Tomas Mikolov, Kai Chen, Greg Corrado, Jeffrey Dean: "Efficient
    Estimation of Word Representations in Vector Space" <https://arxiv.org/abs/1301.3781>`_.
    """

    def __init__(
        self,
        corpus_file=None,
        vector_size=100,
        alpha=0.025,
        window=5,
        min_count=5,
        max_vocab_size=None,
        sample=1e-3,
        seed=1,
        workers=3,
        min_alpha=0.0001,
        sg=0,
        hs=0,
        negative=5,
        ns_exponent=0.75,
        cbow_mean=1,
        hashfxn=hash,
        epochs=5,
        null_word=0,
        trim_rule=None,
        sorted_vocab=1,
        batch_words=10_000,
        compute_loss=False,
        callbacks=(),
        comment=None,
        max_final_vocab=None,
        shrink_windows=True,
        mean_vector=True,
    ):
        """Train, use and evaluate neural networks described in https://code.google.com/p/word2vec/.

        Once you're finished training a model (=no more updates, only querying)
        store and use only the :class:`~gensim.models.keyedvectors.KeyedVectors` instance in ``self.wv``
        to reduce memory.

        The full model can be stored/loaded via its :meth:`~gensim.models.word2vec.Word2Vec.save` and
        :meth:`~gensim.models.word2vec.Word2Vec.load` methods.

        The trained word vectors can also be stored/loaded from a format compatible with the
        original word2vec implementation via `self.wv.save_word2vec_format`
        and :meth:`gensim.models.keyedvectors.KeyedVectors.load_word2vec_format`.

        Parameters
        ----------
        sentences : iterable of iterables, optional
            The `sentences` iterable can be simply a list of lists of tokens, but for larger corpora,
            consider an iterable that streams the sentences directly from disk/network.
            See :class:`~gensim.models.word2vec.BrownCorpus`, :class:`~gensim.models.word2vec.Text8Corpus`
            or :class:`~gensim.models.word2vec.LineSentence` in :mod:`~gensim.models.word2vec` module for such examples.
            See also the `tutorial on data streaming in Python
            <https://rare-technologies.com/data-streaming-in-python-generators-iterators-iterables/>`_.
            If you don't supply `sentences`, the model is left uninitialized -- use if you plan to initialize it
            in some other way.
        corpus_file : str, optional
            Path to a corpus file in :class:`~gensim.models.word2vec.LineSentence` format.
            You may use this argument instead of `sentences` to get performance boost. Only one of `sentences` or
            `corpus_file` arguments need to be passed (or none of them, in that case, the model is left uninitialized).
        vector_size : int, optional
            Dimensionality of the word vectors.
        window : int, optional
            Maximum distance between the current and predicted word within a sentence.
        min_count : int, optional
            Ignores all words with total frequency lower than this.
        workers : int, optional
            Use these many worker threads to train the model (=faster training with multicore machines).
        sg : {0, 1}, optional
            Training algorithm: 1 for skip-gram; otherwise CBOW.
        hs : {0, 1}, optional
            If 1, hierarchical softmax will be used for model training.
            If 0, and `negative` is non-zero, negative sampling will be used.
        negative : int, optional
            If > 0, negative sampling will be used, the int for negative specifies how many "noise words"
            should be drawn (usually between 5-20).
            If set to 0, no negative sampling is used.
        ns_exponent : float, optional
            The exponent used to shape the negative sampling distribution. A value of 1.0 samples exactly in proportion
            to the frequencies, 0.0 samples all words equally, while a negative value samples low-frequency words more
            than high-frequency words. The popular default value of 0.75 was chosen by the original Word2Vec paper.
            More recently, in https://arxiv.org/abs/1804.04212, Caselles-Dupré, Lesaint, & Royo-Letelier suggest that
            other values may perform better for recommendation applications.
        cbow_mean : {0, 1}, optional
            If 0, use the sum of the context word vectors. If 1, use the mean, only applies when cbow is used.
        alpha : float, optional
            The initial learning rate.
        min_alpha : float, optional
            Learning rate will linearly drop to `min_alpha` as training progresses.
        seed : int, optional
            Seed for the random number generator. Initial vectors for each word are seeded with a hash of
            the concatenation of word + `str(seed)`. Note that for a fully deterministically-reproducible run,
            you must also limit the model to a single worker thread (`workers=1`), to eliminate ordering jitter
            from OS thread scheduling. (In Python 3, reproducibility between interpreter launches also requires
            use of the `PYTHONHASHSEED` environment variable to control hash randomization).
        max_vocab_size : int, optional
            Limits the RAM during vocabulary building; if there are more unique
            words than this, then prune the infrequent ones. Every 10 million word types need about 1GB of RAM.
            Set to `None` for no limit.
        max_final_vocab : int, optional
            Limits the vocab to a target vocab size by automatically picking a matching min_count. If the specified
            min_count is more than the calculated min_count, the specified min_count will be used.
            Set to `None` if not required.
        sample : float, optional
            The threshold for configuring which higher-frequency words are randomly downsampled,
            useful range is (0, 1e-5).
        hashfxn : function, optional
            Hash function to use to randomly initialize weights, for increased training reproducibility.
        epochs : int, optional
            Number of iterations (epochs) over the corpus. (Formerly: `iter`)
        trim_rule : function, optional
            Vocabulary trimming rule, specifies whether certain words should remain in the vocabulary,
            be trimmed away, or handled using the default (discard if word count < min_count).
            Can be None (min_count will be used, look to :func:`~gensim.utils.keep_vocab_item`),
            or a callable that accepts parameters (word, count, min_count) and returns either
            :attr:`gensim.utils.RULE_DISCARD`, :attr:`gensim.utils.RULE_KEEP` or :attr:`gensim.utils.RULE_DEFAULT`.
            The rule, if given, is only used to prune vocabulary during build_vocab() and is not stored as part of the
            model.

            The input parameters are of the following types:
                * `word` (str) - the word we are examining
                * `count` (int) - the word's frequency count in the corpus
                * `min_count` (int) - the minimum count threshold.
        sorted_vocab : {0, 1}, optional
            If 1, sort the vocabulary by descending frequency before assigning word indexes.
            See :meth:`~gensim.models.keyedvectors.KeyedVectors.sort_by_descending_frequency()`.
        batch_words : int, optional
            Target size (in words) for batches of examples passed to worker threads (and
            thus cython routines).(Larger batches will be passed if individual
            texts are longer than 10000 words, but the standard cython code truncates to that maximum.)
        compute_loss: bool, optional
            If True, computes and stores loss value which can be retrieved using
            :meth:`~gensim.models.word2vec.Word2Vec.get_latest_training_loss`.
        callbacks : iterable of :class:`~gensim.models.callbacks.CallbackAny2Vec`, optional
            Sequence of callbacks to be executed at specific stages during training.
        shrink_windows : bool, optional
            New in 4.1. Experimental.
            If True, the effective window size is uniformly sampled from  [1, `window`]
            for each target word during training, to match the original word2vec algorithm's
            approximate weighting of context words by distance. Otherwise, the effective
            window size is always fixed to `window` words to either side.
        mean_vector : bool, optional
            Whether to predict a word vector for each individual word, or whether to
            average them.

        Examples
        --------
        Initialize and train a :class:`~gensim.models.word2vec.Word2Vec` model

        .. sourcecode:: pycon

            >>> from gensim.models import Word2Vec
            >>> sentences = [["cat", "say", "meow"], ["dog", "say", "woof"]]
            >>> model = Word2Vec(sentences, min_count=1)

        Attributes
        ----------
        wv : :class:`~gensim.models.keyedvectors.KeyedVectors`
            This object essentially contains the mapping between words and embeddings. After training, it can be used
            directly to query those embeddings in various ways. See the module level docstring for examples.

        """
        self.model = None
        self.corpus_file = corpus_file
        self.vector_size = vector_size
        self.alpha = alpha
        self.window = window
        self.min_count = min_count
        self.max_vocab_size = max_vocab_size
        self.sample = sample
        self.seed = seed
        self.workers = workers
        self.min_alpha = min_alpha
        self.sg = sg
        self.hs = hs
        self.negative = negative
        self.ns_exponent = ns_exponent
        self.cbow_mean = cbow_mean
        self.hashfxn = hashfxn
        self.epochs = epochs
        self.null_word = null_word
        self.trim_rule = trim_rule
        self.sorted_vocab = sorted_vocab
        self.batch_words = batch_words
        self.compute_loss = compute_loss
        self.callbacks = callbacks
        self.comment = comment
        self.max_final_vocab = max_final_vocab
        self.shrink_windows = shrink_windows
        self.mean_vector = mean_vector

    def fit(self, X, y=None):
        """Fit the model according to the given training data.
        Parameters
        ----------
        X : iterable of iterables of str
            The input corpus. X can be simply a list of lists of tokens, but for larger corpora,
            consider an iterable that streams the sentences directly from disk/network.
            See :class:`~gensim.models.word2vec.BrownCorpus`, :class:`~gensim.models.word2vec.Text8Corpus`
            or :class:`~gensim.models.word2vec.LineSentence` in :mod:`~gensim.models.word2vec` module for such examples.
        Returns
        -------
        :class:`~gensim.sklearn_api.w2vmodel.W2VTransformer`
            The trained model.
        """
        self.model = Word2Vec(
            sentences=X,
            corpus_file=self.corpus_file,
            vector_size=self.vector_size,
            alpha=self.alpha,
            window=self.window,
            min_count=self.min_count,
            max_vocab_size=self.max_vocab_size,
            sample=self.sample,
            seed=self.seed,
            workers=self.workers,
            min_alpha=self.min_alpha,
            sg=self.sg,
            hs=self.hs,
            negative=self.negative,
            ns_exponent=self.ns_exponent,
            cbow_mean=self.cbow_mean,
            hashfxn=self.hashfxn,
            epochs=self.epochs,
            null_word=self.null_word,
            trim_rule=self.trim_rule,
            sorted_vocab=self.sorted_vocab,
            batch_words=self.batch_words,
            compute_loss=self.compute_loss,
            callbacks=self.callbacks,
            comment=self.comment,
            max_final_vocab=self.max_final_vocab,
            shrink_windows=self.shrink_windows,
        )
        return self

    def infer_single_token(
        self,
        token: str,
    ) -> Optional[np.ndarray]:
        """
        Use the trained Word2Vec to infer the vector of a single token.
        If the token is unknown it returns an empty numpy array.
        """
        return self.model.wv[token] if token in self.model.wv else np.array([])

    def infer_list_of_tokens(
        self,
        tokens: List[str],
    ) -> List[np.ndarray]:
        """
        If the word vectors of multiple token in a document.
        """
        return [self.infer_single_token(token) for token in tokens if token]

    def mean_of_vector_list(
        self,
        vector_list: List[np.ndarray],
    ) -> np.ndarray:
        """
        Take the mean word vector of a document. It excludes tokens that do
        not have a vector.
        """
        vectors_without_nones = [vector for vector in vector_list if vector.size > 0]
        return np.mean(vectors_without_nones, axis=0)

    def transform(self, X):
        """
        Transform a list of documents. It returns the mean vector for each document.
        Documents that do not have any recognised tokens returns np.nan.
        """
        if self.model is None:
            raise NotFittedError(
                "This model has not been fitted yet. Call 'fit' with appropriate arguments before using this method."
            )

        word_tokens = [self.infer_list_of_tokens(tokens) for tokens in X]

        # Calculate the mean vector of all tokens in each document
        if self.mean_vector:
            word_tokens = [self.mean_of_vector_list(tokens) for tokens in word_tokens]

        return word_tokens

    def partial_fit(self, X):
        raise NotImplementedError(
            "'partial_fit' has not been implemented for W2VTransformer. "
            "However, the model can be updated with a fixed vocabulary using Gensim API call."
        )



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

default_packages = [
    "pandas==1.3.3",
    "dagster==1.0.7",
    "dagster-aws==0.16.7",
    "mlflow==1.20.2",
    "scikit-learn==1.0",
    "pyarrow==6.0.0",
    "markupsafe==2.0.1",
]

remote_packages = [
    "dask==2021.9.1",
    "dask-ml==1.9.0",
    "distributed==2021.9.1",
    "blosc==1.10.6",
    "lz4==3.1.3",
    "s3fs==2021.10.1",
    "psycopg2-binary==2.9.2",
]

test_packages = [
    "pytest==6.2.5",
]

development_packages = [
    "black==21.9b0",
    "isort==5.9.3",
    "setuptools",
    "wheel",
    "twine",
    "ipykernel",
]

nlp_packages = [
    "stop-words==2018.7.23",
    "yake==0.4.8",
    "gensim==4.1.2",
    "spacy==3.2.0",
    "nltk==3.6.5",
]

setup(
    name="quantile-data-kit",
    version="0.0.47",
    author="Jules Huisman",
    author_email="jules.huisman@quantile.nl",
    description="An internal Quantile development kit for making working with data easier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/quantile-development/quantile-data-kit",
    project_urls={
        "Bug Tracker": "https://github.com/quantile-development/quantile-data-kit/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        *default_packages,
        *remote_packages,
        *nlp_packages,
    ],
    extras_require={
        "test": [
            *test_packages,
        ],
        "development": [
            *development_packages,
            *test_packages,
        ],
    },
    packages=find_packages(),
    python_requires=">=3.8",
)



from typing import Union

from dagster import In, Out
from dask.dataframe import DataFrame as DaskDataFrame
from dask.dataframe import Series as DaskSeries
from pandas import DataFrame, Series
from qdk.dagster_types import DataFrameType, MLFlowRunType, SeriesType
from qdk.inference.base import BaseInference
from qdk.mlflow import MLFlowRun


class MLFlowInference(BaseInference):
    compute_function = "predict"
    input_defs = {
        "X": In(DataFrameType),
        "mlflow_run": In(MLFlowRunType),
        "y_true": In(SeriesType, default_value=None),
    }
    output_defs = {"X_hat": Out(DataFrameType)}

    @classmethod
    def predict(
        cls,
        X: Union[DataFrame, DaskDataFrame],
        mlflow_run: MLFlowRun,
        y_true: Union[Series, DaskSeries] = None,
    ) -> Union[DataFrame, DaskDataFrame]:
        # Use the mlflow model to predict
        X_hat: Union[Series, DaskSeries] = mlflow_run.model.predict(X)

        X["y_hat"] = X_hat

        # If y_true is provided, add it to the output
        if y_true is not None:
            X["y_true"] = y_true

        return X



import multiprocessing
from typing import List, Tuple, Union

import dask.dataframe as dd
import pandas as pd
from dagster import Out
from gensim.models.doc2vec import Doc2Vec, TaggedDocument
from pandas.core.frame import DataFrame
from qdk.dagster_types import DataFrameType, ModelType
from qdk.training.word2vec import Word2VecTrainer


class Doc2VecTrainer(Word2VecTrainer):
    output_defs = {
        "model": Out(ModelType),
        "data": Out(DataFrameType, is_required=False),
    }

    @staticmethod
    def tokens_to_tagged_documents(
        token_series: Union[pd.Series, dd.Series],
    ) -> List[TaggedDocument]:
        """Doc2Vec requires TaggedDocuments as input to the model.

        Args:
            token_series (Union[pd.Series, dd.Series]): The input tokens as series.

        Returns:
            List[TaggedDocument]: The tokens encoded as TaggedDocuments
        """
        return [TaggedDocument(doc, [i]) for i, doc in enumerate(token_series)]

    @staticmethod
    def infer_vectors(
        doc2vec_model: Doc2Vec,
        token_series: Union[pd.Series, dd.Series],
    ) -> Union[pd.Series, dd.Series]:
        """Use the Doc2Vec model to infer the embeddings of the token series.

        Args:
            doc2vec_model (Doc2Vec): The Doc2Vec model
            token_series (Union[pd.Series, dd.Series]): The tokens to transform

        Returns:
            Union[pd.Series, dd.Series]: The embeddings
        """
        return token_series.map(doc2vec_model.infer_vector)

    @classmethod
    def train(
        cls,
        X: Union[pd.DataFrame, dd.DataFrame],
        token_column: str = "_tokens_text",
        vector_size: int = 100,
        window: int = 5,
        min_count: int = 5,
        negative: int = 5,
    ) -> Tuple[Doc2Vec, Union[pd.DataFrame, dd.DataFrame]]:
        token_series = X[token_column]

        # Transform the series of tokens to TaggedDocuments
        tagged_documents = cls.tokens_to_tagged_documents(token_series=token_series)

        # Train the Doc2Vec model
        doc2vec_model = Doc2Vec(
            documents=tagged_documents,
            vector_size=vector_size,
            window=window,
            min_count=min_count,
            negative=negative,
            workers=multiprocessing.cpu_count(),
        )

        transformed_column_name = f"_vector_{token_column.replace('_tokens_', '')}"

        X[transformed_column_name] = cls.infer_vectors(
            doc2vec_model=doc2vec_model, token_series=token_series
        )

        return (doc2vec_model, X)



from .dagster_types import DataFrameType, MLFlowRunType, ModelType, SeriesType
from .exporter.base import BaseExporter
from .exporter.database import DatabaseExporter
from .inference.base import BaseInference
from .inference.mlflow import MLFlowInference
from .loader.base import BaseLoader
from .loader.dataframe import DataFrameLoader
from .loader.s3_pickle import S3PickleLoader
from .loader.sql_query import SqlQueryLoader
from .loader.sql_table import SqlTableLoader
from .models.word2vec import Word2VecModel
from .resources.io_manager import qdk_fs_io_manager, qdk_io_manager
from .s3_connection import S3Connection
from .training.base import BaseTrainer
from .training.doc2vec import Doc2VecTrainer
from .training.mlflow import MLFlowTrainingComponent
from .training.sklearn import SklearnTrainer
from .training.word2vec import Word2VecTrainer
from .transform.base import BaseTransformer
from .transform.noun_phrase import NounPhraseTransformer
from .transform.tokenize import TokenizeTransformer
from .transform.train_test import TrainTestTransformer
from .transform.yake import YakeTransformer



from typing import Dict, Union

import dask.dataframe as dd
import pandas as pd
from dagster import Any, In, Out
from qdk.base import BaseComponent


class BaseInference(BaseComponent):
    compute_function = "predict"
    tags = {
        "kind": "inference",
    }
    input_defs = {"data": In(pd.DataFrame), "model": In(Any)}
    output_defs = {"predictions": Out(pd.DataFrame)}

    @classmethod
    def predict(
        cls,
        df: Union[pd.DataFrame, dd.DataFrame],
        model: Any,
    ):
        raise NotImplementedError(
            'Make sure you added a "predict" function to the component'
        )



import os

from dagster import Bool, Field, IOManager, OutputContext, StringSource, check
from dagster import io_manager as dagster_io_manager
from dagster.core.storage.fs_io_manager import PickledObjectFilesystemIOManager
from qdk.materialization import Materializer


class IOManager(IOManager):
    def __init__(self):
        self.values = {}

    def handle_output(self, context: OutputContext, obj):
        keys = context.get_output_identifier()
        self.values[tuple(keys)] = obj

        # Create a materializer for the object
        asset_materialization = Materializer(context, obj).materialize()

        # If the asset was materialized, yield the asset
        if asset_materialization:
            yield asset_materialization

    def load_input(self, context):
        keys = tuple(context.upstream_output.get_output_identifier())

        if keys not in self.values:
            raise KeyError(
                "The key was not found. Make sure the io_manager lives between execution processes."
            )

        obj = self.values[keys]
        return obj


class FilesystemIOManager(PickledObjectFilesystemIOManager):
    def __init__(self, base_dir=None, store_runs=False):
        super().__init__(base_dir=base_dir)
        self.store_runs = check.bool_param(store_runs, "store_runs")

    def _get_path(self, context: OutputContext):
        """Automatically construct filepath."""
        keys = context.get_output_identifier()

        # If not storing individual runs
        # replace the run id with the pipeline name
        if not self.store_runs:
            keys = [context.pipeline_name] + keys[1:]

        return os.path.join(self.base_dir, *keys)

    def handle_output(self, context: OutputContext, obj):
        """
        Pickle the data and store the object to a file.
        We added custom functionality to materialize the asset.
        """
        # Use the inherited handle_output to store the object
        super().handle_output(context, obj)

        # Create a materializer for the object
        asset_materialization = Materializer(context, obj).materialize()

        # If the asset was materialized, yield the asset
        if asset_materialization:
            yield asset_materialization


@dagster_io_manager
def qdk_io_manager(_):
    return IOManager()


@dagster_io_manager(
    config_schema={
        "base_dir": Field(StringSource, is_required=False),
        "store_runs": Field(
            Bool,
            default_value=False,
            description="Whether to create seperate directories for each individual run.",
        ),
    }
)
def qdk_fs_io_manager(init_context):
    base_dir = init_context.resource_config.get(
        "base_dir", init_context.instance.storage_directory()
    )
    store_runs = init_context.resource_config.get("store_runs")
    return FilesystemIOManager(base_dir, store_runs)
